import unittest
import json

from restosaur import API
from restosaur.contrib.apiroot import ApiRoot
from restosaur.contrib.django.dispatch import resource_dispatcher_factory
from .utils import response_content_as_text


class APIRootTestCase(unittest.TestCase):
    def setUp(self):
        super(APIRootTestCase, self).setUp()

        from django.test import RequestFactory

        self.api = API('foo')
        self.apiroot = ApiRoot()
        self.root = self.api.resource('/')
        self.root.get()(self.apiroot.as_view())  # register root view

        self.rqfactory = RequestFactory()

    def call(self, resource, method, *args, **kw):
        rq = getattr(self.rqfactory, method)(resource.path, *args, **kw)
        return resource_dispatcher_factory(self.api, resource)(rq)


class RootPageTestCase(APIRootTestCase):
    def test_successful_returning_status_code_200(self):
        resp = self.call(self.root, 'get')
        self.assertEqual(resp.status_code, 200)

    def test_not_exposing_unregistered_resource(self):
        new_resource = self.api.resource('bar')

        @new_resource.get()
        def bar_get_view(ctx):
            return ctx.Entity()

        resp = self.call(self.root, 'get')
        data = json.loads(response_content_as_text(resp))

        self.assertFalse('bar' in data)

    def test_successful_exposing_registered_resource(self):
        new_resource = self.api.resource('bar')

        @new_resource.get()
        def bar_get_view(ctx):
            return ctx.Entity()

        self.apiroot.register(new_resource, 'bar')

        resp = self.call(self.root, 'get')
        data = json.loads(response_content_as_text(resp))

        self.assertTrue('bar' in data)

    def test_successful_exposing_registered_resource_with_automatic_name(self):
        new_resource = self.api.resource('bar')

        @new_resource.get()
        def bar_get_view(ctx):
            return ctx.Entity()

        self.apiroot.register(new_resource)

        resp = self.call(self.root, 'get')
        data = json.loads(response_content_as_text(resp))

        self.assertTrue('bar' in data)

    def test_exposing_registered_resource_with_valid_uri(self):
        new_resource = self.api.resource('bar')

        @new_resource.get()
        def bar_get_view(ctx):
            return ctx.Entity()

        self.apiroot.register(new_resource)

        resp = self.call(self.root, 'get')
        data = json.loads(response_content_as_text(resp))

        self.assertEqual(data['bar'], 'http://testserver/foo/bar')

    def test_exposing_registered_resource_with_valid_uri_with_parameter(self):
        new_resource = self.api.resource('bar/:id')

        @new_resource.get()
        def bar_get_view(ctx):
            return ctx.Entity()

        self.apiroot.register(new_resource)

        resp = self.call(self.root, 'get')
        data = json.loads(response_content_as_text(resp))

        self.assertEqual(data['bar'], 'http://testserver/foo/bar/:id')


class ApiRootRootResourceRegistrationTestCase(unittest.TestCase):
    def setUp(self):
        super(ApiRootRootResourceRegistrationTestCase, self).setUp()

        from django.test import RequestFactory

        self.api = API('foo')
        self.root_resource = self.api.resource('/')
        self.rqfactory = RequestFactory()

    def call(self, resource, method, *args, **kw):
        rq = getattr(self.rqfactory, method)(resource.path, *args, **kw)
        return resource_dispatcher_factory(self.api, resource)(rq)

    def test_successful_registration_as_a_GET_method(self):
        ApiRoot(self.root_resource)
        self.assertTrue(self.root_resource.is_callback_registered('GET'))

    def test_successful_registration_proper_apiroot_func(self):
        apiroot = ApiRoot(self.root_resource)
        rootfuncname = apiroot.as_view().__name__
        self.assertEqual(
                self.root_resource.get_callback('GET').__name__,
                rootfuncname)

    def test_exception_at_second_registration_apiroot_to_same_rootresource(self):
        ApiRoot(self.root_resource)
        with self.assertRaises(ValueError):
            ApiRoot(self.root_resource)

    def test_that_apiroot_entries_added_after_registration_are_accessible(self):
        apiroot = ApiRoot(self.root_resource)
        bar = self.api.resource('bar')
        apiroot.register(bar)

        resp = self.call(self.root_resource, 'get')
        data = json.loads(response_content_as_text(resp))

        self.assertTrue('bar' in data)



import functools
import logging
import sys
import urllib

from collections import OrderedDict, defaultdict

from .exceptions import Http404
from .representations import (
        RepresentationAlreadyRegistered, ValidatorAlreadyRegistered,
        Representation, Validator)
from .utils import join_content_type_with_vnd, split_mediatype
from . import contentnegotiation, responses, urltemplate, serializers


log = logging.getLogger(__name__)


def _join_ct_vnd(content_type, vnd):
    return join_content_type_with_vnd(content_type, vnd)


def dict_as_text(obj, ctx, depth=0):
    output = u''

    for key, value in obj.items():
        if isinstance(value, dict):
            value = u'\n'+dict_as_text(value, ctx, depth+1)
        output += u'%s%s: %s\n' % (' '*depth*2, key, value)

    return output


def resource_name_from_path(path):
    return urltemplate.remove_parameters(path).strip('/')


class NoMoreMediaTypes(Exception):
    pass


class NoRepresentationFound(Exception):
    pass


class Resource(object):
    def __init__(
            self, api, path, name=None,
            default_content_type='application/json',
            link_model=None, link_name=None):
        self._api = api
        self._path = path
        self._required_parameters = urltemplate.get_parameters(self._path)
        self._callbacks = defaultdict(dict)
        self._registered_methods = set()
        self._name = name or resource_name_from_path(path)
        self._representations = OrderedDict()
        self._validators = OrderedDict()
        self._default_content_type = default_content_type
        self._supported_media_types = defaultdict(set)

        # register "pass-through" validators

        for content_type, serializer in serializers.get_all():
            self.add_validator(content_type=content_type)

        if link_model:
            self._api.register_view(
                    model=link_model, resource=self, view_name=link_name)

        # register aliases for the decorators
        for verb in ('GET', 'POST', 'PUT', 'DELETE', 'OPTIONS', 'PATCH'):
            setattr(
                self, verb.lower(), functools.partial(self._decorator, verb))

    def is_callback_registered(self, method, content_type=None):
        content_type = content_type or self._default_content_type
        return method in self._callbacks[content_type]

    def register_method_callback(self, callback, method, content_type=None):
        content_type = content_type or self._default_content_type
        self._callbacks[content_type][method] = callback
        self._registered_methods.add(method)
        self._supported_media_types[method].add(content_type)

    def get_method_supported_mediatypes(self, method):
        return list(self._supported_media_types[method])

    def get_callback(self, method, content_type=None):
        content_type = content_type or self._default_content_type
        return self._callbacks[content_type][method]

    def get_allowed_methods(self):
        return list(self._registered_methods)

    @property
    def default_content_type(self):
        return self._default_content_type

    def _decorator(self, method, accept=None):
        def wrapper(view):
            if self.is_callback_registered(method, content_type=accept):
                raise ValueError(
                        'Method `%s` is already registered' % method)
            self.register_method_callback(
                    view, method=method, content_type=accept)
            return view
        return wrapper

    def _match_media_type(self, accept, representations, exclude=None):
        exclude = exclude or []

        def _drop_mt_args(x):
            return x.split(';')[0]

        mediatypes = list(filter(
                lambda x: _drop_mt_args(x) not in exclude, map(
                    lambda x: x.media_type(), representations)))

        if not mediatypes:
            raise NoMoreMediaTypes

        mediatype = contentnegotiation.best_match(mediatypes, accept)

        if not mediatype:
            raise NoMoreMediaTypes

        return _drop_mt_args(mediatype)

    def _match_representation(self, instance, ctx, accept=None):

        # Use "*/*" as default -- RFC 7231 (Section 5.3.2)
        accept = accept or ctx.headers.get('accept') or '*/*'

        exclude = []
        model = type(instance)

        representations = self.representations

        while True:
            try:
                mediatype = self._match_media_type(
                        accept, representations, exclude=exclude)
            except NoMoreMediaTypes:
                break

            if not self.has_representation_for(model, mediatype):
                exclude.append(mediatype)
            else:
                return self.get_representation(model, mediatype)

        for mediatype in exclude:
            if self.has_representation_for(None, mediatype):
                return self.get_representation(None, mediatype)

        raise NoRepresentationFound(
            '%s has no registered representation handler for `%s`' % (
                model, accept))

    def _http_response(self, response):
        content = ''
        content_type = self._default_content_type

        if response.data is not None:
            try:
                representation = self._match_representation(
                        response.data, response.context)
            except NoRepresentationFound:
                if isinstance(response, responses.SuccessfulResponse):
                    return self._http_response(
                        responses.NotAcceptableResponse(
                            response.context, headers=response.headers))
                elif isinstance(response, (
                        responses.ServerErrorResponse,
                        responses.ClientErrorResponse)):
                    # For errors use any representation supported by
                    # server. It is better to provide any information
                    # in any format instead of nothing.
                    # -- RFC7231 (Section 6.5 & 6.6)
                    #
                    # Restosaur will try to match most acceptible
                    # representation.

                    accept = response.context.headers.get('accept')
                    accepting = ['*/*;q=0.1']

                    if accept:
                        media_type, media_subtype = accept.split('/')
                        accepting.insert(0, '%s/*;q=1' % media_type)

                    try:
                        representation = self._match_representation(
                                response.data, response.context,
                                accept=','.join(accepting))
                    except NoRepresentationFound:
                        # return no content and preserve status code
                        pass
                    else:
                        content = representation.render(
                                response.context, response.data)
                        content_type = _join_ct_vnd(
                               representation.content_type, representation.vnd)
                else:
                    # return no content and preserve status code
                    pass
            else:
                content = representation.render(
                        response.context, response.data)
                content_type = _join_ct_vnd(
                       representation.content_type, representation.vnd)

        return (response, content, content_type)

    @property
    def name(self):
        return self._name

    @property
    def api(self):
        return self._api

    @property
    def path(self):
        return self._path

    @property
    def representations(self):
        result = []
        for models in self._representations.values():
            result += models.values()
        return result + self._api.representations

    def has_representation_for(self, model, media_type):
        return (media_type in self._representations
                and model in self._representations[media_type]) or (
                        self._api.has_representation_for(model, media_type))

    def get_representation(self, model, media_type):
        try:
            return self._representations[media_type][model]
        except KeyError:
            return self._api.get_representation(model, media_type)

    def model(self, view_name=None):
        """
        Decorator for registering `self` (the resource)
        as a view for the model
        """
        def register_model(model_class):
            self._api.register_view(
                    model=model_class, resource=self, view_name=view_name)
            return model_class
        return register_model

    def __call__(self, ctx, *args, **kw):
        from django.http import Http404 as DjangoHttp404

        method = ctx.method
        request = ctx.request

        # support for X-HTTP-METHOD-OVERRIDE

        method = ctx.headers.get('x-http-method-override') or method

        # Check request method and raise MethodNotAllowed if unsupported

        allowed_methods = self.get_allowed_methods()

        if method not in allowed_methods:
            headers = {
                    'Allow': ', '.join(allowed_methods),
                }
            return self._http_response(ctx.MethodNotAllowed({
                'error': 'Method `%s` is not registered for resource `%s`' % (
                    method, self._path)}, headers=headers))

        # Negotiate payload content type and store the best matching
        # result in ctx.request_content_type

        if ctx.content_type and ctx.content_length:
            media_types = self.get_method_supported_mediatypes(method)
            if media_types:
                ctx.request_content_type = contentnegotiation.best_match(
                        media_types, ctx.content_type)
            else:
                # server does not support any representation
                ctx.request_content_type = None
        elif ctx.content_length:
            # No payload content-type was provided.
            # According to RFC7231 (Section 3.1.1.5) server may assume
            # "application/octet-stream" or try to examine the type.

            ctx.request_content_type = 'application/octet-stream'
        else:
            # No payload
            ctx.request_content_type = None

        # match response representation, serializer and content type

        if ctx.content_length and ctx.content_type:
            if ctx.request_content_type:
                if request.body:
                    try:
                        ctx.validator = self._validators[
                                    ctx.request_content_type]
                    except KeyError:
                        pass
                    else:
                        try:
                            ctx.body = ctx.validator.parse(ctx)
                        except serializers.DeserializationError as ex:
                            resp = responses.exception_response_factory(
                                    ctx, ex, cls=responses.BadRequestResponse)
                            return self._http_response(resp)
            elif not ctx.content_length:
                ctx.body = None
            else:
                return self._http_response(ctx.UnsupportedMediaType())

        log.debug('Calling %s, %s, %s' % (method, args, kw))

        try:
            callback = self.get_callback(method, ctx.request_content_type)

            try:
                resp = callback(ctx, *args, **kw)
            except DjangoHttp404:
                raise Http404
            else:
                if not resp:
                    raise TypeError(
                            'Function `%s` does not return '
                            'a response object' % callback)
                return self._http_response(resp)
        except Http404:
            return self._http_response(ctx.NotFound())
        except Exception as ex:
            if self._api.debug:
                tb = sys.exc_info()[2]
            else:
                tb = None
            resp = responses.exception_response_factory(ctx, ex, tb)
            log.exception(
                    'Internal Server Error: %s', ctx.request.path,
                    exc_info=sys.exc_info(),
                    extra={
                        'status_code': resp.status,
                        'context': ctx,
                    }
            )
            return self._http_response(resp)

    def accept(self, media_type=None):
        media_type = media_type or self._default_content_type
        ct, vnd = split_mediatype(media_type)
        return self.validator(content_type=ct, vnd=vnd)

    def representation(self, model=None, media=None, serializer=None):
        def wrapped(func):
            if isinstance(media, (list, tuple)):
                content_types = map(split_mediatype, media)
            else:
                content_types = [split_mediatype(
                    media or self._default_content_type)]

            for ct, v, args in content_types:
                self.add_representation(
                    model=model, vnd=v, content_type=ct, qvalue=args.get('q'),
                    serializer=serializer, _transform_func=func)
            return func
        return wrapped

    def validator(self, vnd=None, content_type=None, serializer=None):
        def wrapped(func):
            self.add_validator(
                    vnd=vnd, content_type=content_type, serializer=serializer,
                    _validator_func=func)
            return func
        return wrapped

    def add_representation(
            self, model=None, vnd=None, content_type=None, qvalue=None,
            serializer=None, _transform_func=None):

        content_type = content_type or self._default_content_type
        repr_key = _join_ct_vnd(content_type, vnd)

        if (repr_key in self._representations and
                not repr_key == self._default_content_type
                and model in self._representations[repr_key]):
            raise RepresentationAlreadyRegistered(
                    '%s: %s (%s)' % (self._path, repr_key, model))

        obj = Representation(
                vnd=vnd, content_type=content_type, serializer=serializer,
                _transform_func=_transform_func, qvalue=qvalue)

        self._representations.setdefault(repr_key, {})
        self._representations[repr_key][model] = obj
        return obj

    def add_validator(
            self, vnd=None, content_type=None, serializer=None,
            _validator_func=None):

        content_type = content_type or self._default_content_type
        repr_key = _join_ct_vnd(content_type, vnd)

        if (repr_key in self._validators and
                not repr_key == self._default_content_type):
            raise ValidatorAlreadyRegistered(
                    '%s: %s' % (self._path, repr_key))

        obj = Validator(
                vnd=vnd, content_type=content_type, serializer=serializer,
                _validator_func=_validator_func)

        self._validators[content_type] = obj
        return obj

    def uri(self, context, params=None, query=None):
        assert params is None or isinstance(
                params, dict), "entity.uri() params should be passed as dict"

        params = params or {}

        uri = context.build_absolute_uri(self._path)
        uri = urltemplate.to_url(uri, params)

        if query:
            uri += '?'+urllib.urlencode(query)

        return uri



import unittest

import django

from django.test import SimpleTestCase, Client


class DjangoAppBasedAutodiscoverTestCase(SimpleTestCase):
    @unittest.skipIf(
            django.VERSION < (1, 7, 0),
            'Not supported for Django %s' % django.get_version())
    def test_successful_autoregistering_root_resource_from_app(self):
        django.setup()

        with self.settings(
                ROOT_URLCONF='tests.urls_autoregister',
                INSTALLED_APPS=['restosaur', 'tests.artest']):
            from .urls_autoregister import api
            self.assertEqual(len(api.resources), 1)
            self.assertEqual(api.resources[0].path, '/')

    @unittest.skipIf(
            django.VERSION < (1, 7, 0),
            'Not supported for Django %s' % django.get_version())
    def test_valid_response_from_autoregistered_app(self):
        c = Client()

        with self.settings(
                ROOT_URLCONF='tests.urls_autoregister',
                INSTALLED_APPS=['restosaur', 'tests.artest']):
            resp = c.get('/')
            self.assertEqual(resp.status_code, 200)

    @unittest.skipIf(
            django.VERSION < (1, 7, 0),
            'Not supported for Django %s' % django.get_version())
    def test_successful_registering_artest_app(self):
        from django.apps import apps
        django.setup()

        with self.settings(
                ROOT_URLCONF='tests.urls_autoregister',
                INSTALLED_APPS=['restosaur', 'tests.artest']):
            self.assertTrue(apps.is_installed('tests.artest'))



import json
import unittest

from restosaur import API
from restosaur.contrib.django import API as DjangoAPI
from restosaur.contrib.django.dispatch import resource_dispatcher_factory
from .utils import response_content_as_text


class APITestCase(unittest.TestCase):
    def setUp(self):
        from django.test import RequestFactory

        super(APITestCase, self).setUp()
        self.rqfactory = RequestFactory()

    def call(self, api, resource, method, *args, **kw):
        rq = getattr(self.rqfactory, method)(resource.path, *args, **kw)
        return resource_dispatcher_factory(api, resource)(rq)


class APIPathsTestCase(APITestCase):

    def test_appending_slash_to_api_path(self):
        api = API('foo')
        self.assertEqual(api.path, 'foo/')

    def test_not_appending_slash_to_api_path_if_exists(self):
        api = API('foo/')
        self.assertEqual(api.path, 'foo/')

    def test_that_root_url_pattern_does_not_contain_slash(self):
        api = DjangoAPI('foo')
        api.resource('/')
        urls = api.get_urls()
        root_url = urls[0].url_patterns[0]
        self.assertFalse('/' in root_url._regex)

    def test_that_typical_url_pattern_does_not_contain_prepending_slash(self):
        api = DjangoAPI('foo')
        api.resource('bar/')
        urls = api.get_urls()
        bar_url = urls[0].url_patterns[0]
        self.assertEqual(bar_url._regex, '^bar/$')

    def test_status_200OK_of_nonprefixed_api_path(self):
        api = API()
        root = api.resource('/')

        @root.get()
        def root_view(ctx):
            return ctx.Response({'root': 'ok'})

        resp = self.call(api, root, 'get')
        self.assertEqual(resp.status_code, 200)

    def test_valid_response_of_nonprefixed_api_path(self):
        api = API()
        root = api.resource('/')

        @root.get()
        def root_view(ctx):
            return ctx.Response({'root': 'ok'})

        resp = self.call(api, root, 'get')
        resp_json = json.loads(response_content_as_text(resp))
        self.assertEqual(resp_json['root'], 'ok')



# coding: utf-8
from __future__ import unicode_literals

import datetime
import six
import unittest

from restosaur import API, responses
from restosaur.contrib.django.dispatch import build_context


class ContextTestCase(unittest.TestCase):
    def setUp(self):
        from django.test import RequestFactory

        super(ContextTestCase, self).setUp()

        def create_context(method, path, resource, api=None, **kwargs):
            api = api or self.api
            request = getattr(self.rqfactory, method)(path)
            return api.make_context(
                    host=request.get_host(), path=request.path,
                    request=request, resource=resource,
                    method=method, **kwargs)

        self.api = API('/')
        self.api2 = API('webapi/')
        self.api3 = API('webapi')
        self.rqfactory = RequestFactory()
        self.factory = create_context
        self.ctx = create_context('get', '', lambda ctx: None, api=self.api)
        self.ctx2 = create_context('get', '', lambda ctx: None, api=self.api2)
        self.ctx3 = create_context('get', '', lambda ctx: None, api=self.api3)

    def test_deserialized_property(self):
        ctx = self.factory('get', '/test/', lambda ctx: None, body='body')
        self.assertEqual(ctx.body, ctx.deserialized)


class TestContextBuilidURI(ContextTestCase):
    def test_using_current_location(self):
        ctx = self.factory('get', '/test/', lambda ctx: None)
        self.assertEqual(ctx.build_absolute_uri(), 'http://testserver/test/')

    def test_using_custom_location(self):
        ctx = self.factory('get', '/foo/', lambda ctx: None)
        self.assertEqual(
                ctx.build_absolute_uri('/bar/'), 'http://testserver/bar/')

    def test_appending_custom_parameter_to_current_location(self):
        ctx = self.factory('get', '/foo/', lambda ctx: None)
        self.assertEqual(
                ctx.build_absolute_uri(parameters={'bar': 'baz'}),
                'http://testserver/foo/?bar=baz')

    def test_appending_custom_parameter_to_custom_location(self):
        ctx = self.factory('get', '/foo/', lambda ctx: None)
        url = ctx.build_absolute_uri('/bar/', parameters={'spam': 'eggs'})
        self.assertEqual(url, 'http://testserver/bar/?spam=eggs')

    def test_overriding_parameter_on_current_location(self):
        ctx = self.factory('get', '/foo/?bar=spam', lambda ctx: None)
        url = ctx.build_absolute_uri(parameters={'bar': 'eggs'})
        self.assertEqual(url, 'http://testserver/foo/?bar=eggs')

    def test_overriding_parameter_on_custom_location(self):
        ctx = self.factory('get', '/foo/?bar=spam', lambda ctx: None)
        url = ctx.build_absolute_uri('/baz/', parameters={'bar': 'eggs'})
        self.assertEqual(url, 'http://testserver/baz/?bar=eggs')

    def test_encoding_parameter(self):
        ctx = self.factory('get', '/foo/', lambda ctx: None)
        url = ctx.build_absolute_uri('/baz/', parameters={
            'bar': six.text_type('zażółć')
        })
        self.assertEqual(
                url, 'http://testserver/baz/?bar=za%C5%BC%C3%B3%C5%82%C4%87')

    def test_generating_absolute_uri_using_request_path(self):
        ctx = self.factory('get', '/', lambda ctx: None)
        self.assertEqual(ctx.build_absolute_uri(), 'http://testserver/')

    def test_generating_absolute_uri_using_slash_as_path(self):
        ctx = self.factory('get', '', lambda ctx: None)
        self.assertEqual(ctx.build_absolute_uri('/'), 'http://testserver/')

    def test_generating_absolute_uri_using_double_slash_as_path(self):
        ctx = self.factory('get', '', lambda ctx: None)
        self.assertEqual(ctx.build_absolute_uri('//'), 'http://testserver/')

    def test_generating_absolute_uri_using_triple_slash_as_path(self):
        ctx = self.factory('get', '', lambda ctx: None)
        self.assertEqual(ctx.build_absolute_uri('///'), 'http://testserver/')

    def test_generating_absolute_uri_using_path_with_one_preceding_slash(self):
        ctx = self.factory('get', '', lambda ctx: None)
        self.assertEqual(
                ctx.build_absolute_uri('/some'), 'http://testserver/some')

    def test_generating_absolute_uri_using_path_with_third_preceding_slashes(self):  # NOQA
        ctx = self.factory('get', '', lambda ctx: None)
        self.assertEqual(
                ctx.build_absolute_uri('///some'), 'http://testserver/some')

    def test_generating_absolute_uri_using_path_ending_with_slash(self):
        ctx = self.factory('get', '', lambda ctx: None)
        self.assertEqual(
                ctx.build_absolute_uri('/some/'), 'http://testserver/some/')

    def test_not_replacing_hostname_when_using_two_prepending_slashes_in_build_absolute_uri(self):  # NOQA
        ctx = self.factory('get', '', lambda ctx: None)
        self.assertEqual(
                ctx.build_absolute_uri('//some/'), 'http://testserver/some/')

    def test_successful_generating_uri_using_shortcut_wrapper_to_resource(self):  # NOQA
        res = self.api.resource('foo/:pk')
        ctx = self.factory('get', '/bar/', lambda ctx: None)
        url = ctx.url_for(res, pk='test')
        self.assertEqual(url, 'http://testserver/foo/test')

    def test_generating_uri_for_prefixed_api_and_slash_as_path(self):
        self.assertEqual(
                self.ctx2.build_absolute_uri('/'),
                'http://testserver/webapi/')

    def test_generating_uri_for_prefixed_api_and_path_wihout_preceding_slash(self):  # NOQA
        self.assertEqual(
                self.ctx2.build_absolute_uri('foo'),
                'http://testserver/webapi/foo')

    def test_generating_uri_for_prefixed_api_and_path_wihout_preceding_slash_but_ending_with_slash(self):  # NOQA
        self.assertEqual(
                self.ctx2.build_absolute_uri('foo/'),
                'http://testserver/webapi/foo/')

    def test_generating_uri_for_prefixed_api_without_slash_and_slash_as_path(self):  # NOQA
        self.assertEqual(
                self.ctx3.build_absolute_uri('/'),
                'http://testserver/webapi/')

    def test_generating_uri_for_prefixed_api_without_slash_and_path_wihout_preceding_slash(self):  # NOQA
        self.assertEqual(
                self.ctx3.build_absolute_uri('foo'),
                'http://testserver/webapi/foo')

    def test_generating_uri_for_prefixed_api_without_slash_and_path_wihout_preceding_slash_but_ending_with_slash(self):  # NOQA
        self.assertEqual(
                self.ctx3.build_absolute_uri('foo/'),
                'http://testserver/webapi/foo/')

    def test_successful_building_uri_with_multivalue_query_arg(self):
        ctx = self.factory('get', '', lambda ctx: None)
        uri = ctx.build_absolute_uri('/some', {'foo': ['bar', 'baz']})
        self.assertEqual(uri, 'http://testserver/some?foo=bar&foo=baz')

    def test_successful_getting_mutlivalue_GET_args(self):
        res = self.api.resource('foo')
        rq = self.rqfactory.get('/foo?bar=baz&bar=qux')
        ctx = build_context(self.api, res, rq)
        self.assertTrue('baz' in ctx.parameters['bar'])
        self.assertTrue('qux' in ctx.parameters['bar'])


class TestContextIfModifiedSince(ContextTestCase):
    def test_succesful_parse_rfc_1123_example_time(self):
        from restosaur.context import parse_http_date
        dt = parse_http_date('foo', {'foo': 'Tue, 15 Nov 1994 08:12:31 GMT'})
        self.assertEqual(dt, datetime.datetime(1994, 11, 15, 8, 12, 31))

    def test_error_parse_rfc_1123(self):
        from restosaur.context import parse_http_date
        dt = parse_http_date('foo', {'foo': 'invalid http date'})
        self.assertIsNone(dt)

    def test_that_date_is_modified(self):
        ctx = self.factory('get', '/foo/', lambda ctx: None, headers={
            'if-modified-since': 'Tue, 15 Nov 1994 08:12:31 GMT',
            })
        modification_date = datetime.datetime(1995, 1, 1, 10, 0, 0)
        self.assertTrue(ctx.is_modified_since(modification_date))

    def test_that_date_is_modified_for_invalid_header(self):
        ctx = self.factory('get', '/foo/', lambda ctx: None, headers={
            'if-modified-since': 'invalid http header',
            })
        modification_date = datetime.datetime(1995, 1, 1, 10, 0, 0)
        self.assertTrue(ctx.is_modified_since(modification_date))

    def test_that_date_is_modified_without_header(self):
        ctx = self.factory('get', '/foo/', lambda ctx: None)
        modification_date = datetime.datetime(1995, 1, 1, 10, 0, 0)
        self.assertTrue(ctx.is_modified_since(modification_date))

    def test_that_date_is_not_modified(self):
        ctx = self.factory('get', '/foo/', lambda ctx: None, headers={
            'if-modified-since': 'Tue, 15 Nov 1994 08:12:31 GMT',
            })
        modification_date = datetime.datetime(1990, 1, 1, 10, 0, 0)
        self.assertFalse(ctx.is_modified_since(modification_date))


class TestContextResponseFactories(ContextTestCase):
    def setUp(self):
        super(TestContextResponseFactories, self).setUp()
        self.ctx = self.factory('get', '/foo/', lambda ctx: None)

    def test_base_response(self):
        obj = self.ctx.Response()
        self.assertTrue(isinstance(obj, responses.Response))

    def test_created_response(self):
        obj = self.ctx.Created()
        self.assertTrue(isinstance(obj, responses.CreatedResponse))

    def test_nocontent_response(self):
        obj = self.ctx.NoContent()
        self.assertTrue(isinstance(obj, responses.NoContentResponse))

    def test_unauthorized_response(self):
        obj = self.ctx.Unauthorized()
        self.assertTrue(isinstance(obj, responses.UnauthorizedResponse))

    def test_forbidden_response(self):
        obj = self.ctx.Forbidden()
        self.assertTrue(isinstance(obj, responses.ForbiddenResponse))

    def test_notfound_response(self):
        obj = self.ctx.NotFound()
        self.assertTrue(isinstance(obj, responses.NotFoundResponse))

    def test_methodnotallowed_response(self):
        obj = self.ctx.MethodNotAllowed()
        self.assertTrue(isinstance(obj, responses.MethodNotAllowedResponse))

    def test_notacceptable_response(self):
        obj = self.ctx.NotAcceptable()
        self.assertTrue(isinstance(obj, responses.NotAcceptableResponse))

    def test_entity_response(self):
        obj = self.ctx.Entity()
        self.assertTrue(isinstance(obj, responses.EntityResponse))

    def test_collection_response(self):
        obj = self.ctx.Collection([])
        self.assertTrue(isinstance(obj, responses.CollectionResponse))

    def test_validationerror_response(self):
        obj = self.ctx.ValidationError({})
        self.assertTrue(isinstance(obj, responses.ValidationErrorResponse))

    def test_notmodified_response(self):
        obj = self.ctx.NotModified()
        self.assertTrue(isinstance(obj, responses.NotModifiedResponse))

    def test_seeother_response(self):
        obj = self.ctx.SeeOther('/')
        self.assertTrue(isinstance(obj, responses.SeeOtherResponse))



import json
from django.test import SimpleTestCase, Client
from .utils import response_content_as_text


class URLPatternsTestCase(SimpleTestCase):
    def test_successful_calling_apiroot_without_path_prefix(self):
        c = Client()

        with self.settings(ROOT_URLCONF='tests.urls_root_without_prefix'):
            resp = c.get('/')
            self.assertEqual(resp.status_code, 200)

    def test_valid_response_from_apiroot_without_path_prefix(self):
        c = Client()

        with self.settings(ROOT_URLCONF='tests.urls_root_without_prefix'):
            resp = c.get('/')
            resp_json = json.loads(response_content_as_text(resp))
            self.assertEqual(resp_json['root'], 'ok')

    def test_successful_calling_apiroot_with_path_prefix(self):
        c = Client()

        with self.settings(ROOT_URLCONF='tests.urls_root_with_prefix'):
            resp = c.get('/api/')
            self.assertEqual(resp.status_code, 200)

    def test_that_api_trailing_prefix_does_not_affect_urlresolver(self):
        c = Client()

        with self.settings(ROOT_URLCONF='tests.urls_root_with_trailing_slash'):
            resp = c.get('/api/')
            self.assertEqual(resp.status_code, 200)



from collections import defaultdict

import six

from .representations import (
        RepresentationAlreadyRegistered, UnknownRepresentation,
        Representation, RestosaurExceptionDict,
        restosaur_exception_dict_as_text)
from .resource import Resource
from .utils import autodiscover, join_content_type_with_vnd
from .loading import load_resource
from .context import Context


class ModelViewAlreadyRegistered(Exception):
    pass


class ModelViewNotRegistered(Exception):
    pass


class BaseAPI(object):
    def __init__(
            self, path=None, middlewares=None,
            context_class=None, default_charset=None, debug=False):
        path = path or ''
        if path and not path.endswith('/'):
            path += '/'
        if path and path.startswith('/'):
            path = path[1:]
        self.path = path
        self.debug = debug
        self.resources = []
        self.default_charset = default_charset or 'utf-8'
        self.middlewares = middlewares or []
        self._representations = defaultdict(dict)  # type->repr_key
        self._model_views = defaultdict(dict)
        self.context_class = context_class or Context

    def make_context(self, **kwargs):
        return self.context_class(self, **kwargs)

    def add_resources(self, *resources):
        self.resources += resources

    def resource(self, path, *args, **kw):
        obj = Resource(self, path, *args, **kw)
        self.add_resources(obj)
        return obj

    def add_representation(
            self, type_, content_type, vnd=None, qvalue=None,
            serializer=None, _transform_func=None):

        representation = Representation(
            content_type=content_type, vnd=vnd,
            serializer=serializer, _transform_func=_transform_func,
            qvalue=qvalue)

        self.register_representation(type_, representation)

    def register_representation(self, type_, representation):

        content_type = representation.content_type
        vnd = representation.vnd
        repr_key = join_content_type_with_vnd(content_type, vnd)

        if (repr_key in self._representations and
                type_ in self._representations[repr_key]):
            raise RepresentationAlreadyRegistered(
                            '%s for %s' % (repr_key, type_))

        self._representations[repr_key][type_] = representation

    def get_representation(self, model, media_type):
        try:
            return self._representations[media_type][model]
        except KeyError:
            raise UnknownRepresentation('%s for %s' % (
                        media_type, model))

    def has_representation_for(self, model, media_type):
        return (media_type in self._representations
                and model in self._representations[media_type])

    @property
    def representations(self):
        result = []
        for models in self._representations.values():
            result += models.values()
        return result

    def resource_for_viewmodel(self, model, view_name=None):
        try:
            model_meta = self._model_views[model]
        except KeyError:
            raise ModelViewNotRegistered(
                    'No views are registered for %s' % model)

        try:
            resource = model_meta[view_name]
        except KeyError:
            if not view_name:
                raise ModelViewNotRegistered(
                    'No default view registered for %s' % model)
            else:
                raise ModelViewNotRegistered(
                    'View `%s` is not registered for %s' % (view_name, model))
        else:
            if isinstance(resource, six.string_types):
                resource = load_resource(resource)
                model_meta[view_name] = resource
            return resource

    def register_view(self, model, resource, view_name=None):
        if view_name in self._model_views[model]:
            if view_name:
                raise ModelViewAlreadyRegistered(
                    '%s is already registered as a "%s" view' % (
                        model, view_name))
            else:
                raise ModelViewAlreadyRegistered(
                    '%s is already registered as a default view' % model)

        self._model_views[model][view_name] = resource

    def view(self, resource, view_name=None):
        """
        A shortcut decorator for registering a `model_class`
        as as a `view_name` view of a `resource`.

        The `resource` may be passed as a dotted string path
        to avoid circular import problems.
        """

        def register_view_for_model(model_class):
            self.register_view(
                model_class, resource=resource, view_name=view_name)
            return model_class
        return register_view_for_model

    def autodiscover(self, *args, **kw):
        """
        Shortcut for `restosaur.autodiscover()`
        """
        autodiscover(*args, **kw)


class API(BaseAPI):
    def __init__(self, *args, **kw):
        super(API, self).__init__(*args, **kw)

        # backward compatibility
        configure_plain_text_api(self)
        configure_json_api(self)


JSON = API


def configure_json_api(api):
    api.add_representation(
            RestosaurExceptionDict, content_type='application/json')
    api.add_representation(
            dict, content_type='application/json')

    # backward compatibility

    from .utils import Collection
    api.add_representation(
            Collection, content_type='application/json')


def configure_plain_text_api(api):
    api.add_representation(
            RestosaurExceptionDict, content_type='text/plain',
            _transform_func=restosaur_exception_dict_as_text,
            qvalue=0.1)


def api_factory(path=None, api_class=API, **kwargs):
    api = api_class(path, **kwargs)
    return api


def json_api_factory(path=None, api_class=API, **kwargs):
    api = api_factory(path=path, api_class=api_class, **kwargs)
    return api



from __future__ import absolute_import

from django.conf import settings
from django.utils.encoding import force_text
from django.views import debug

from ...api import API as BaseAPI
from ...representations import RestosaurExceptionDict


def django_html_exception(obj, ctx):
    if settings.DEBUG:
        resp = debug.technical_500_response(
                ctx.request, exc_type=obj.exc_type,
                exc_value=obj.exc_value,
                tb=obj.tb, status_code=obj.status_code)
        return force_text(resp.content)
    return '<h1>Internal Server Error (%s)</h1>' % obj.status_code


class API(BaseAPI):
    def __init__(self, *args, **kw):
        from django.conf import settings

        charset = kw.pop('default_charset', None) or settings.DEFAULT_CHARSET
        debug = kw.pop('debug', settings.DEBUG)

        self.append_slash = kw.pop('append_slash', False)

        kw['default_charset'] = charset
        kw['debug'] = debug

        super(API, self).__init__(*args, **kw)

        self.add_representation(
                RestosaurExceptionDict, content_type='text/html',
                _transform_func=django_html_exception)

    def get_urls(self):
        try:
            from django.conf.urls import patterns, url, include
        except ImportError:
            from django.conf.urls import url, include

            def patterns(x, *urls):
                return list(urls)

        from django.views.decorators.csrf import csrf_exempt
        from .dispatch import resource_dispatcher_factory
        from ... import urltemplate

        urls = []

        for resource in self.resources:
            path = urltemplate.to_django_urlpattern(resource._path)

            if path.startswith('/'):
                path = path[1:]

            if not path.endswith('/') and self.append_slash:
                path = path+'/'

            urls.append(url(
                '^%s$' % path, csrf_exempt(
                    resource_dispatcher_factory(self, resource))))

        return [url('^%s' % self.path, include(patterns('', *urls)))]

    def urlpatterns(self):
        try:
            from django.conf.urls import patterns, include
        except ImportError:
            return self.get_urls()
        else:
            return patterns('', (r'^', include(self.get_urls())))



from . import contentnegotiation


def model_to_dict(obj, context):
    """
    Convert django model instance to dict
    """

    data = {}

    for field in obj._meta.fields:
        field_name = field.column if field.rel else field.name
        data[field.column] = getattr(obj, field_name)

    return data


class Collection(object):
    def __init__(
            self, context, iterable, key=None, totalcount=None,
            totalcount_key=None):

        self.totalcount = totalcount or len(iterable)
        self.iterable = iterable
        self.totalcount_key = totalcount_key or 'totalCount'
        self.key = key or 'items'


def join_content_type_with_vnd(content_type, vnd):
    if not vnd:
        return content_type

    x, y = content_type.split('/')
    return '%s/%s+%s' % (x, vnd, y)


def autodiscover(module_name='restapi'):
    from django.conf import settings

    try:
        from django.utils.module_loading import autodiscover_modules
    except ImportError:
        from django.utils.importlib import import_module
        from django.utils.module_loading import module_has_submodule
        autodiscover_modules = None

    if autodiscover_modules:
        autodiscover_modules(module_name)
    else:
        for app in settings.INSTALLED_APPS:
            mod = import_module(app)
            try:
                import_module('%s.%s' % (app, module_name))
            except:
                if module_has_submodule(mod, module_name):
                    raise


def split_mediatype(mt):
    """
    Split media type into (content_type, vnd, params) tuple

    Function converts `type/[tree.]subtype[+suffix][;params]`
    into generic mediatype together with subtype without suffix.
    Extra parameteres are returned as third element of the tuple.
    """

    type_, subtype, args = contentnegotiation.parse_media_type(mt)
    suffix = subtype.split('+')[-1] if '+' in subtype else None

    if suffix and suffix not in (
            'json', 'zip', 'ber', 'der', 'fastinfoset', 'wbxml'):
        raise ValueError('Suffix "%s" is not supported (see: RFC6839)')

    ct = '%s/%s' % (type_, suffix or subtype)
    vnd = subtype.split('+')[0] if suffix else None
    return ct, vnd, args


def generic_mediatype(mt):
    return split_mediatype(mt)[0]



import unittest
import json

from restosaur import API
from restosaur.contrib.django.dispatch import resource_dispatcher_factory
from .utils import response_content_as_text


class BaseTestCase(unittest.TestCase):
    def setUp(self):
        super(BaseTestCase, self).setUp()

        from django.test import RequestFactory

        self.api = API('foo')
        self.collection = self.api.resource('items')
        self.detail = self.api.resource('items/detail')

        @self.detail.get()
        def detail_ctrl(ctx):
            return ctx.OK({'bar': 'baz'})

        @self.detail.post('application/json')
        def detail_post_json(ctx):
            return ctx.OK(ctx.body)

        @self.detail.post('text/plain')
        def detail_post_textplain(ctx):
            return ctx.OK({'bar': ctx.body})

        @self.detail.post('test/error500')
        def detail_post_error500(ctx):
            raise RuntimeError('a runtime error')

        self.rqfactory = RequestFactory()

    def call(self, resource, method, *args, **kw):
        rq = getattr(self.rqfactory, method)(resource.path, *args, **kw)
        return resource_dispatcher_factory(self.api, resource)(rq)


class ContentNegotiationTestCase(BaseTestCase):
    def test_not_acceptable_plaintext_GET(self):
        resp = self.call(self.detail, 'get', HTTP_ACCEPT='text/plain')
        self.assertEqual(resp.status_code, 406)

    def test_successful_plaintext_GET_for_explicite_representation(self):
        @self.detail.representation(media='text/plain')
        def detail_text_plain(obj, ctx):
            return str(obj)
        resp = self.call(self.detail, 'get', HTTP_ACCEPT='text/plain')
        self.assertEqual(resp.content, b"{'bar': 'baz'}")

    def test_successful_json_GET_using_default_representation(self):
        resp = self.call(self.detail, 'get', HTTP_ACCEPT='application/json')
        self.assertEqual(resp.status_code, 200)


class ErrorsContentNegotiationTestCase(BaseTestCase):
    def test_returning_plaintext_error_message(self):
        resp = self.call(
                self.detail, 'post', content_type='test/error500',
                HTTP_ACCEPT='text/plain')
        self.assertIn(b'RuntimeError: a runtime error', resp.content)

    def test_returning_json_error_message(self):
        resp = self.call(
                self.detail, 'post', content_type='test/error500',
                HTTP_ACCEPT='application/json')
        data = json.loads(response_content_as_text(resp))
        self.assertEqual(data['error'], 'a runtime error')

    def test_returning_json_error_message_for_not_accepted_content_type(self):
        resp = self.call(
                self.detail, 'post', content_type='test/error500',
                HTTP_ACCEPT='foo/bar')
        self.assertEqual(resp['Content-Type'], 'application/json')

    def test_returning_plaintex_error_message_for_not_accepted_text_type(self):
        resp = self.call(
                self.detail, 'post', content_type='test/error500',
                HTTP_ACCEPT='text/unsupported')
        self.assertEqual(resp['Content-Type'], 'text/plain')


class DeprecatedResponseTestCase(BaseTestCase):
    def test_returning_406_when_used_base_Response_class(self):
        @self.collection.get()
        def collection_ctrl(ctx):
            return ctx.Response({'items': [{'bar': 'baz'}]})
        resp = self.call(
                self.collection, 'get', HTTP_ACCEPT='text/unsupported')
        self.assertEqual(resp.status_code, 406)



import traceback

from .utils import Collection, join_content_type_with_vnd
from . import serializers


class RepresentationAlreadyRegistered(Exception):
    pass


class UnknownRepresentation(Exception):
    pass


class ValidatorAlreadyRegistered(Exception):
    pass


def _pass_through_trasnform(x, ctx):
    return x


def _pass_through_validation(x, ctx):
    return x


class Representation(object):
    def __init__(
            self, vnd=None, content_type='application/json', qvalue=None,
            serializer=None, _transform_func=None):

        self.qvalue = qvalue if qvalue is not None else 1
        self.serializer = serializer or serializers.get(content_type)
        self.content_type = content_type
        self.vnd = vnd
        self._transform_func = _transform_func or _pass_through_trasnform

    def render(self, context, obj):
        """
        Renders representation of `obj` as raw content
        """
        if isinstance(obj, Collection):
            items = map(
                lambda x: self._transform_func(x, context), obj.iterable)
            data = {
                    obj.key: list(items),
                    obj.totalcount_key: len(obj.iterable),
                    }
        else:
            data = self._transform_func(obj, context)
        return self.serializer.dumps(data)

    def media_type(self):
        mt = join_content_type_with_vnd(self.content_type, self.vnd)
        return '%s; q=%s' % (mt, self.qvalue)


class Validator(object):
    def __init__(
            self, vnd=None, content_type='application/json',
            serializer=None, _validator_func=None):

        self.serializer = serializer or serializers.get(content_type)
        self.content_type = content_type
        self.vnd = vnd
        self._validator_func = _validator_func or _pass_through_validation

    def parse(self, context):
        """
        Parses raw representation content and builds object
        """
        return self._validator_func(self.serializer.loads(context), context)


class RestosaurExceptionDict(dict):
    def __init__(self, ex, tb=None):
        super(RestosaurExceptionDict, self).__init__()
        self.exc_type = type(ex)
        self.exc_value = ex
        self.tb = tb
        self.status_code = 500


def restosaur_exception_dict_as_text(obj, ctx):
    output = u'\n'.join(
            traceback.format_exception(obj.exc_type, obj.exc_value, obj.tb))

    return output



from django.apps import AppConfig


class RestosaurAppConfig(AppConfig):
    name = 'restosaur'
    verbose_name = 'Restosaur'

    def ready(self):
        from .settings import AUTODISCOVER_MODULE, AUTODISCOVER
        from . import autodiscover

        if AUTODISCOVER:
            autodiscover(AUTODISCOVER_MODULE)



def resource_dispatcher_factory(api, resource, response_builder, context_builder):
    def dispatch_request(request, *args, **kw):
        ctx = context_builder(api, resource, request)
        bypass_resource_call = False
        middlewares_called = []

        for middleware in api.middlewares:
            middlewares_called.append(middleware)

            try:
                method = middleware.process_request
            except AttributeError:
                pass
            else:
                if method(request, ctx) is False:
                    bypass_resource_call = True
                    break

        if not bypass_resource_call:
            response, content, ct = resource(ctx, *args, **kw)
        else:
            response, content, ct = None, None, None

        middlewares_called.reverse()

        for middleware in middlewares_called:
            try:
                method = middleware.process_response
            except AttributeError:
                pass
            else:
                if method(request, response, ctx) is False:
                    break

        return response_builder(response, content, ct)
    return dispatch_request



from . import _mimeparsepatched


def best_match(acceptables, header):
    """
    Match "Accept" `header` together with `acceptables` defined on
    the server side
    """

    return _mimeparsepatched.best_match(acceptables, header)


def parse_media_type(media_type):
    return _mimeparsepatched.parse_mime_type(media_type)



from .. import urltemplate


class ResourceAlreadyRegistered(Exception):
    pass


def autogenerate_resource_name(resource):
    path = urltemplate.remove_parameters(resource.path)

    try:
        name = list(filter(None, path.split('/')))[-1]
    except IndexError:
        name = '/'

    return name


class ApiRoot(object):
    def __init__(self, root_resource=None):
        """
        Creates instance of ApiRoot registry.

        If `root_resource` is provided, the ApiRoot`s view will be registered
        as a root_resource`s service for HTTP GET method.
        """

        self.resources = {}

        if root_resource:
            root_resource.get()(self.as_view())

    def register(self, resource, name=None):
        name = name or autogenerate_resource_name(resource)

        if name in self.resources:
            raise ResourceAlreadyRegistered(name)
        self.resources[name] = resource

    def expose(self, name, resource):
        def wrap(resource):
            self.register(resource, name)
            return resource
        return wrap

    def as_view(self):
        def get_api_root(ctx):
            data = {}

            for name, resource in self.resources.items():
                data[name] = resource.uri(ctx)

            return ctx.OK(data)
        return get_api_root



from django.http import HttpResponse

from restosaur.context import QueryDict
from restosaur.headers import normalize_header_name
from restosaur import dispatch as restosaur_dispatch


def build_context(api, resource, request):
    try:
        # Django may raise RawPostDataException sometimes;
        # i.e. when processing POST multipart/form-data;
        # In that cases we can't access raw body anymore, sorry

        raw_body = request.body
    except:
        raw_body = None

    parameters = {}

    if request.resolver_match:
        parameters.update(request.resolver_match.kwargs)

    parameters.update(QueryDict(list(request.GET.lists())))

    headers = dict(map(
        lambda x: (normalize_header_name(x[0]), x[1]),
        filter(lambda x: x[0].startswith('HTTP_'), request.META.items())))

    try:
        content_length = int(request.META['CONTENT_LENGTH'])
    except (KeyError, TypeError, ValueError):
        content_length = 0

    content_type = request.META.get('CONTENT_TYPE')

    return api.make_context(
            host=request.get_host(), path=request.path,
            method=request.method, parameters=parameters,
            data=request.POST, files=request.FILES, raw=raw_body,
            charset=request.encoding or api.default_charset,
            secure=request.is_secure(), encoding=request.GET.encoding,
            resource=resource, request=request, headers=headers,
            content_type=content_type, content_length=content_length)


def _do_http_response(response, content, content_type):
    """
    RESTResponse -> HTTPResponse factory
    """

    if isinstance(response, HttpResponse):
        return response

    httpresp = HttpResponse(content, status=response.status)

    if content_type:
        httpresp['Content-Type'] = content_type

    for header, value in response.headers.items():
        httpresp[header] = value

    return httpresp


def response_builder(response, content, content_type):
    if response is None:
        return HttpResponse()
    else:
        return _do_http_response(response, content, content_type)


def resource_dispatcher_factory(api, resource):
    return restosaur_dispatch.resource_dispatcher_factory(
                    api, resource, response_builder, build_context)



import collections
import email
import functools
import six

try:
    import urlparse
except ImportError:
    from urllib import parse as urlparse

try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode

import times
# todo: implement own conversion utility
from django.utils.encoding import force_bytes

from .loading import load_resource
from . import responses


def parse_http_date(header, headers):
    if header in headers and headers[header]:
        timetuple = email.utils.parsedate_tz(headers[header])
        try:
            return times.from_unix(email.utils.mktime_tz(timetuple))
        except (TypeError, ValueError):
            pass


class QueryDict(collections.MutableMapping):
    """
    QueryDict acts like a plain `dict` type, but it handles
    automatially multiple values for same key.

    The most safest representation of URI query parameters is a list
    of tuples, because the parameter names aren't unique. Unfortunately
    accessing list of tuples is not so handy, so a mapping is
    required.

    In most cases query parameters looks like a mapping of simple
    key => value pairs, so we're expecting just one value per key. But when
    value is a list, we're expecting that accessing a key will return that
    list, not last nor first value.

    The problematic case is for keys, for which we're expecting always a list
    of values, but just one was passed in URI. Accessing the key will give
    just straight value instead of expected list with one item. In that cases
    you should use `QueryDict.getlist()` directly, which returns always a list.

    The values are stored internally as lists.

    `.items()` method returns a list of (key, value) tuples, where value is
    a single value from a key's values list. This means that key may not be
    unique. This representation is compatible with `urllib.urlencode()`.

    `.keys()` returns unique key names, same as for pure `dict`.

    `.values()` returns list of same values, which can be accessed by key,

    `.lists()` returns internal representation as list of lists.
    """

    def __init__(self, initial=None):
        self._data = {}
        self.update(initial)

    def update(self, data):
        if data is None:
            return
        else:
            try:
                data = data.items()
            except AttributeError:
                pass
        keys = set([x[0] for x in data])
        for key in keys:
            self._data[key] = []
        for key, value in data:
            if isinstance(value, (list, tuple)):
                for x in value:
                    self._data[key].append(x)
            else:
                self._data[key].append(value)

    def items(self):
        result = []
        for key, values in self._data.items():
            result += list(map(lambda x: (key, x), values))
        return result

    def getlist(self, key, default=None):
        return self._data.get(key, default)

    def lists(self):
        return self._data.items()

    def __setitem__(self, key, value):
        return self.update({key: value})

    def __getitem__(self, key):
        return self._data[key][-1]\
                if len(self._data[key]) < 2 else self._data[key]

    def __delitem__(self, key):
        del self._data[key]

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __repr__(self):
        return repr(self._data)


class Context(object):
    def __init__(
            self, api, host='localhost', path='/', method='GET',
            parameters=None, body=None, data=None, files=None, raw=None,
            extra=None, headers=None, charset=None, secure=False,
            encoding='utf-8', resource=None, request=None, content_length=None,
            content_type=None):
        self.method = method
        self.api = api
        self.charset = charset
        self.headers = headers or {}
        self.resource = resource
        self.request = request
        self.encoding = encoding
        self.secure = secure
        self.host = host
        self.path = path
        self.body = body
        self.raw = raw
        self.parameters = QueryDict(parameters)  # GET
        self.data = data or {}  # POST
        self.files = files or {}  # FILES
        self.deserializer = None
        self.content_type = content_type
        self.content_length = content_length
        self.extra = extra or {}

    def build_absolute_uri(self, path=None, parameters=None):
        """
        Returns absolute uri to the specified `path` with optional
        query string `parameters`.

        If no `path` is provided, the current request full path
        (including query string) will be used and extended by
        optional `parameters`.
        """

        def build_uri(path):
            current = 'http%s://%s%s' % (
                    's' if self.secure else '', self.host, self.path)
            return urlparse.urljoin(current, path)

        params = QueryDict()
        if path:
            full_path = u'/'.join(
                    filter(None, (self.api.path+path).split('/')))
            if path.endswith('/'):
                full_path += '/'
            uri = build_uri('/'+full_path)
        else:
            params.update(self.parameters.items())
            uri = build_uri(self.path)

        # todo: change to internal restosaur settings
        enc = self.encoding

        params.update(parameters or {})
        params = list(map(
                lambda x: (x[0], force_bytes(x[1], enc)),
                params.items()))

        if params:
            return '%s?%s' % (uri, urlencode(params))
        else:
            return uri

    def url_for(self, resource, **kwargs):
        """
        Shortcut wrapper of `resource.uri()`
        """
        if isinstance(resource, six.string_types):
            resource = load_resource(resource)
        return resource.uri(self, params=kwargs)

    def link_model(self, model_instance, view_name=None, query=None):
        resource = self.api.resource_for_viewmodel(
                type(model_instance), view_name)
        return self.link(
                resource, model=model_instance, query=query)

    def link(self, resource, model=None, query=None):
        """
        Generate URL for `model` instance based on `resource`
        path template.

        Handy shortcut for `resource.uri()`
        """

        if not resource._required_parameters or not model:
            return resource.uri(self, query=query)

        params = {}

        for parameter in resource._required_parameters:
            try:
                params[parameter] = getattr(model, parameter)
            except AttributeError:
                try:
                    params[parameter] = model[parameter]
                except (KeyError, TypeError, ValueError):
                    raise ValueError(
                        'Can\'t construct URL parameter "%s" from `%s`' % (
                            parameter, model))

        return resource.uri(self, params=params, query=query)

    def is_modified_since(self, dt):
        """
        Compares datetime `dt` with `If-Modified-Since` header value.
        Returns True if `dt` is newer than `If-Modified-Since`,
        False otherwise.
        """
        if_modified_since = parse_http_date('if-modified-since', self.headers)

        if if_modified_since:
            return times.to_unix(
                dt.replace(microsecond=0)) > times.to_unix(if_modified_since)

        return True

    @property
    def deserialized(self):
        return self.body

    def wrap(self, func):
        """Wrap `func` with the Context instance as a last argument"""

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            args = list(args)+[self]
            return func(*args, **kwargs)
        return wrapped

    # response factories

    def Continue(self, *args, **kwargs):  # 100
        return responses.ContinueResponse(self, *args, **kwargs)

    def OK(self, *args, **kwargs):  # 200
        return responses.OKResponse(self, *args, **kwargs)

    def Response(self, *args, **kwargs):  # deprecated 200-like response
        return self.OK(*args, **kwargs)

    def Created(self, *args, **kwargs):  # 201
        return responses.CreatedResponse(self, *args, **kwargs)

    def Accepted(self, *args, **kwargs):  # 202
        return responses.AcceptedResponse(self, *args, **kwargs)

    def NoContent(self, *args, **kwargs):  # 204
        return responses.NoContentResponse(self, *args, **kwargs)

    def MovedPermanently(self, *args, **kwargs):  # 301
        return responses.MovedPermanentlyResponse(self, *args, **kwargs)

    def Found(self, *args, **kwargs):  # 302
        return responses.FoundResponse(self, *args, **kwargs)

    def SeeOther(self, *args, **kwargs):  # 303
        return responses.SeeOtherResponse(self, *args, **kwargs)

    def NotModified(self, *args, **kwargs):  # 304
        return responses.NotModifiedResponse(self, *args, **kwargs)

    def BadRequest(self, *args, **kwargs):  # 400
        return responses.BadRequestResponse(self, *args, **kwargs)

    def Unauthorized(self, *args, **kwargs):  # 401
        return responses.UnauthorizedResponse(self, *args, **kwargs)

    def Forbidden(self, *args, **kwargs):  # 403
        return responses.ForbiddenResponse(self, *args, **kwargs)

    def NotFound(self, *args, **kwargs):  # 404
        return responses.NotFoundResponse(self, *args, **kwargs)

    def MethodNotAllowed(self, *args, **kwargs):  # 405
        return responses.MethodNotAllowedResponse(self, *args, **kwargs)

    def NotAcceptable(self, *args, **kwargs):  # 406
        return responses.NotAcceptableResponse(self, *args, **kwargs)

    def Conflict(self, *args, **kwargs):  # 409
        return responses.ConflictResponse(self, *args, **kwargs)

    def Gone(self, *args, **kwargs):  # 410
        return responses.GoneResponse(self, *args, **kwargs)

    def UnsupportedMediaType(self, *args, **kwargs):  # 415
        return responses.UnsupportedMediaTypeResponse(self, *args, **kwargs)

    def ValidationError(self, *args, **kwargs):  # 422 WEBDAV Deprecated
        return responses.ValidationErrorResponse(self, *args, **kwargs)

    def Entity(self, *args, **kwargs):  # deprecated, 200
        return responses.EntityResponse(self, *args, **kwargs)

    def Collection(self, *args, **kwargs):  # deprecated, 200
        return responses.CollectionResponse(self, *args, **kwargs)
